{+--------------------------------------------------------------------------+
 | Unit:        mwPCharList
 | Created:     10.97 - 4.98
 | Author:      Martin Waldenburg
 | Copyright    1997, all rights reserved.
 | Description: TLongIntList is a dynamic array of LongInts.
 |              TPCharList alows you to use a PChar as a list of Strings.
 | Version:     1.2
 | DISCLAIMER:  This is provided as is, expressly without a warranty of any kind.
 |              You use it at your own risc.
 +--------------------------------------------------------------------------+}

unit mwPCharList;

{$R-}

interface

uses SysUtils, Windows;

type

{ TLongIntList class }

  PLongIntArray = ^TLongIntArray;
  TLongIntArray = array[0..0] of LongInt;

  TLongIntList = class(TObject)
  private
    FCapacity: Integer;
    FCount: Integer;
    FLongIntList: PLongIntArray;
  protected
    function GetItems(Index: Integer): LongInt;
    procedure SetCapacity(NewCapacity: Integer);
    procedure SetCount(NewCount: Integer);
    procedure SetItems(Index: Integer; Item: LongInt);
  public
    constructor Create;
    destructor Destroy; override;
    function Add(Item: LongInt): Integer;
    procedure Clear;
    procedure Delete(Index: Integer);
    procedure Exchange(Index1, Index2: Integer);
    function First: LongInt;
    function IndexOf(Item: LongInt): Integer;
    procedure Insert(Index: Integer; Item: LongInt);
    function Last: LongInt;
    procedure Move(CurIndex, NewIndex: Integer);
    function Remove(Item: LongInt): Integer;
    procedure DeleteGroup(StartIndex: LongInt; GroupCount: LongInt);
    property Capacity: Integer read FCapacity write SetCapacity;
    property Count: Integer read FCount write SetCount;
    property Items[Index: Integer]: LongInt read GetItems write SetItems; default;
    property LongIntList: PLongIntArray read FLongIntList;
  end;

{ TPCharList class }

  TPCharList = class(TObject)
  private
    FLinePos: TLongIntList;
    fOrigin: PChar;
    fPCharSize: Longint;
    fPCharCapacity: Longint;
    fPosition: Longint;
    procedure WriteTo(InsPos, DelPos: LongInt; Item: String);
    function GetCount: Integer;
    procedure SetCount(value: Integer);
    function GetCapacity: Integer;
    procedure SetOrigin(value: PChar);
    procedure ResetPositionsFrom(Index, Value: LongInt);
    procedure SetPosition(value: Longint);
  protected
    function GetItems(Index: Integer): String;
    procedure SetCapacity(NewCapacity: Integer);
    procedure SetItems(Index: Integer; Item: String);
  public
    constructor Create;
    destructor Destroy; override;
    function Add(Item: String): Integer;
    procedure Clear;
    procedure Delete(Index: Integer);
    procedure Exchange(Index1, Index2: Integer);
    function First: String;
    function IndexOf(Item: String): Integer;
    procedure Insert(Index: Integer; Item: String);
    function Last: String;
    procedure Move(CurIndex, NewIndex: Integer);
    function Remove(Item: String): Integer;
    procedure ScanForLines;
    function ReadChars(CharCount: LongInt): String;
    procedure WriteString(aString: String);
    procedure DeleteChars(CharCount: LongInt);
    function InsertString(StartIndex: LongInt; ToInsert: String): LongInt;
    procedure DeleteGroup(StartIndex: LongInt; GroupCount: LongInt);
    function MoveGroup(OldStartIndex: LongInt; NewStartIndex: LongInt; GroupCount: LongInt): Boolean;
    property Capacity: Integer read GetCapacity write SetCapacity;
    property Count: Integer read GetCount write SetCount;
    property Items[Index: Integer]: String read GetItems write SetItems; default;
    property LinePosList: TLongIntList read FLinePos;
    property Origin: PChar read fOrigin write SetOrigin;
    property PCharSize: Longint read fPCharSize;
    property PCharCapacity: Longint read fPCharCapacity;
    property Position: Longint read fPosition write SetPosition;
  published
  end;

implementation


{ TLongIntList }

{ TLongIntList constructor and destructor +++++++++++++++++++++++++++++++++++++}

constructor TLongIntList.Create;
begin
  inherited Create;
end; { Create }

destructor TLongIntList.Destroy;
begin
  Clear;
  inherited Destroy;
end; { Destroy }

{ TLongIntList protected ++++++++++++++++++++++++++++++++++++++++++++++++++++++}

function TLongIntList.GetItems(Index: Integer): LongInt;
begin
  Result := FLongIntList[Index];
end; { GetItems }

procedure TLongIntList.SetCapacity(NewCapacity: Integer);
begin
  if NewCapacity < FCount then FCount := NewCapacity;
  if NewCapacity <> FCapacity then
  begin
    ReallocMem(FLongIntList, NewCapacity * SizeOf(LongInt));
    FCapacity := NewCapacity;
  end;
end; { SetCapacity }

procedure TLongIntList.SetCount(NewCount: Integer);
begin
  if NewCount > FCapacity then SetCapacity(NewCount);
  FCount := NewCount;
end; { SetCount }

procedure TLongIntList.SetItems(Index: Integer; Item: LongInt);
begin
  FLongIntList[Index] := Item;
end; { SetItems }

{ TLongIntList public +++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

function TLongIntList.Add(Item: LongInt): Integer;
begin
  Result := FCount;
  if Result + 1 >= FCapacity then SetCapacity(FCapacity + 1024);
  FLongIntList[Result] := Item;
  Inc(FCount);
end; { Add }

procedure TLongIntList.Clear;
begin
  SetCount(0);
  SetCapacity(0);
end; { Clear }

procedure TLongIntList.Delete(Index: Integer);
begin
  Dec(FCount);
  if Index < FCount then
    System.Move(FLongIntList[Index + 1], FLongIntList[Index],
      (FCount - Index) * SizeOf(LongInt));
end; { Delete }

procedure TLongIntList.DeleteGroup(StartIndex: LongInt; GroupCount: LongInt);
begin
  Dec(FCount, GroupCount);
  if StartIndex < FCount then
    System.Move(FLongIntList[StartIndex + GroupCount], FLongIntList[StartIndex],
      (FCount - StartIndex) * SizeOf(LongInt));
end; { DeleteGroup }

procedure TLongIntList.Exchange(Index1, Index2: Integer);
var
  Item: LongInt;
begin
  Item := FLongIntList[Index1];
  FLongIntList[Index1] := FLongIntList[Index2];
  FLongIntList[Index2] := Item;
end; { Exchange }

function TLongIntList.First: LongInt;
begin
  Result := GetItems(0);
end; { First }

function TLongIntList.IndexOf(Item: LongInt): Integer;
begin
  Result := 0;
  while (Result < FCount) and (FLongIntList[Result] <> Item) do Inc(Result);
  if Result = FCount then Result := -1;
end; { IndexOf }

procedure TLongIntList.Insert(Index: Integer; Item: LongInt);
begin
  if FCount = FCapacity then SetCapacity(FCapacity + 1024);
  if Index < FCount then
    System.Move(FLongIntList[Index], FLongIntList[Index + 1],
      (FCount - Index) * SizeOf(LongInt));
  FLongIntList[Index] := Item;
  Inc(FCount);
end; { Insert }

function TLongIntList.Last: LongInt;
begin
  Result := GetItems(FCount - 1);
end; { Last }

procedure TLongIntList.Move(CurIndex, NewIndex: Integer);
var
  Item: LongInt;
begin
  if CurIndex <> NewIndex then
  begin
    Item := GetItems(CurIndex);
    Delete(CurIndex);
    Insert(NewIndex, Item);
  end;
end; { Move }

function TLongIntList.Remove(Item: LongInt): Integer;
begin
  Result := IndexOf(Item);
  if Result <> -1 then Delete(Result);
end; { Remove }

{ TPCharList }

{ TPCharList constructor and destructor +++++++++++++++++++++++++++++++++++++}

constructor TPCharList.Create;
begin
  inherited Create;
  FLinePos := TLongIntList.Create;
  FLinePos.Add(0);
  FPosition := 0;
end; { Create }

destructor TPCharList.Destroy;
begin
  FLinePos.Free;
end; { Destroy }

{ TPCharList private ++++++++++++++++++++++++++++++++++++++++++++++++++++++}

procedure TPCharList.SetOrigin(value: PChar);
begin
  FOrigin := Value;
  if Origin = nil then fPCharSize := 0
  else fPCharSize := StrLen(FOrigin);
  fPCharCapacity := fPCharSize;
  ScanForLines;
end; { SetOrigin }

procedure TPCharList.WriteTo(InsPos, DelPos: LongInt;
  Item: String);
var
  StringCount, NewSize: Longint;
  aString: String;
begin
  aString := Item + (FOrigin + DelPos);
  StringCount := Length(aString);
  if (InsPos >= 0) and (StringCount >= 0) then
  begin
    NewSize := InsPos + StringCount;
    if NewSize > 0 then
    begin
      if NewSize >= FPCharCapacity then
      begin
        Try
          FPCharCapacity := NewSize + 4096;
          ReAllocMem(FOrigin, PCharCapacity);
        except
          raise exception.Create('unable to reallocate PChar');
        end;
      end;
      StrECopy((FOrigin + InsPos), PChar(aString));
      FPCharSize := NewSize;
      FOrigin[FPCharSize] := #0;
      aString := '';
    end;
  end;
end; { WriteTo }

function TPCharList.GetCount: Integer;
begin
  Result := FLinePos.Count - 1;
end; { GetCount }

procedure TPCharList.SetCount(value: Integer);
begin
  FLinePos.Count := Value + 1;
end; { SetCount }

function TPCharList.GetCapacity: Integer;
begin
  Result := FLinePos.Capacity;
end; { GetCapacity }

procedure TPCharList.ResetPositionsFrom(Index, Value: LongInt);
begin
  while Index < FLinePos.Count do
  begin
    FLinePos[Index] := FLinePos[Index] + Value;
    inc(Index);
  end
end; { ResetPositionsFrom }

procedure TPCharList.SetPosition(value: Longint);
begin
  if Value < 0 then FPosition := 0;
  if Value > PCharSize then FPosition := PCharSize;
end; { SetPosition }

{ TPCharList protected ++++++++++++++++++++++++++++++++++++++++++++++++++++++}

function TPCharList.GetItems(Index: Integer): String;
var
  StartPos, EndPos, StringLen: LongInt;
begin
  StartPos := FLinePos[Index];
  EndPos := FLinePos[Index + 1];
  while fOrigin[EndPos - 1] in [#10, #13] do dec(EndPos);
  StringLen := EndPos - StartPos;
  SetString(Result, (FOrigin + StartPos), StringLen);
end; { GetItems }

procedure TPCharList.SetCapacity(NewCapacity: Integer);
begin
  FLinePos.Capacity := NewCapacity;
end; { SetCapacity }

procedure TPCharList.SetItems(Index: Integer; Item: String);
var
  StartPos, EndPos, OldLen, NewLen, Diff: LongInt;
begin
  Item := Item + #13#10;
  StartPos := FLinePos[Index];
  EndPos := FLinePos[Index + 1];
  OldLen := EndPos - StartPos;
  NewLen := Length(Item);
  Diff := NewLen - OldLen;
  WriteTo(StartPos, EndPos, Item);
  ResetPositionsFrom(Index + 1, Diff);
end; { SetItems }

{ TPCharList public +++++++++++++++++++++++++++++++++++++++++++++++++++++++++}

function TPCharList.Add(Item: String): Integer;
var
  StartPos, EndPos: LongInt;
begin
  Result := Count;
  Item := Item + #13#10;
  StartPos := FLinePos[Result];
  EndPos := StartPos + Length(Item);
  FLinePos.Add(EndPos);
  WriteTo(StartPos, StartPos, Item);
end; { Add }

procedure TPCharList.Clear;
begin
  SetCount(0);
  FLinePos.Capacity := 1;
end; { Clear }

procedure TPCharList.Delete(Index: Integer);
var
  StartPos, EndPos, OldLen: LongInt;
begin
  StartPos := FLinePos[Index];
  EndPos := FLinePos[Index + 1];
  OldLen := EndPos - StartPos;
  WriteTo(StartPos, EndPos, '');
  FLinePos.Delete(Index);
  ResetPositionsFrom(Index, -OldLen);
end; { Delete }

procedure TPCharList.DeleteGroup(StartIndex: LongInt; GroupCount: LongInt);
var
  I, StartPos, EndPos, OldLen: LongInt;
begin
  StartPos := FLinePos[StartIndex];
  EndPos := FLinePos[StartIndex + GroupCount];
  OldLen := EndPos - StartPos;
  WriteTo(StartPos, EndPos, '');
  FLinePos.DeleteGroup(StartIndex, GroupCount);
  ResetPositionsFrom(StartIndex, -OldLen);
end; { DeleteGroup }

procedure TPCharList.Exchange(Index1, Index2: Integer);
var
  Item1, Item2: String;
begin
  Item1 := GetItems(Index1);
  Item2 := GetItems(Index2);
  SetItems(Index2, Item1);
  SetItems(Index1, Item2);
end; { Exchange }

function TPCharList.First: String;
begin
  Result := GetItems(0);
end; { First }

function TPCharList.IndexOf(Item: String): Integer;
begin
  Result := 0;
  while (Result < Count) and (GetItems(Result) <> Item) do Inc(Result);
  if Result = Count then Result := -1;
end; { IndexOf }

procedure TPCharList.Insert(Index: Integer; Item: String);
var
  StartPos, EndPos, ItemLen: LongInt;
begin
  Item := Item + #13#10;
  ItemLen := Length(Item);
  StartPos := FLinePos[Index];
  EndPos := StartPos + ItemLen;
  WriteTo(StartPos, StartPos, Item);
  ResetPositionsFrom(Index + 1, ItemLen);
  FLinePos.Insert(Index + 1, EndPos);
end; { Insert }

function TPCharList.InsertString(StartIndex: LongInt; ToInsert: String): LongInt;
var
  I, StartPos, EndPos, ItemLen: LongInt;
  Helper: TPCharList;
begin
  Helper := TPCharList.Create;
  Helper.Add(ToInsert);
  Helper.ScanForLines;
  Result := Helper.Count;
  ToInsert := ToInsert + #13#10;
  ItemLen := Length(ToInsert);
  StartPos := FLinePos[StartIndex];
  EndPos := StartPos + ItemLen;
  WriteTo(StartPos, StartPos, ToInsert);
  ResetPositionsFrom(StartIndex + 1, ItemLen);
  for I := 0 to Helper.Count - 1 do
  begin
    StartPos := FLinePos[StartIndex + I];
    EndPos := StartPos + Length(Helper[I]) + 2;
    FLinePos.Insert(StartIndex + I + 1, EndPos);
  end;
  Helper.Free;
end; { InsertString }

function TPCharList.Last: String;
begin
  Result := GetItems(Count - 1);
end; { Last }

procedure TPCharList.Move(CurIndex, NewIndex: Integer);
var
  Item: String;
begin
  if CurIndex <> NewIndex then
  begin
    Item := GetItems(CurIndex);
    Delete(CurIndex);
    Insert(NewIndex, Item);
  end;
end; { Move }

function TPCharList.MoveGroup(OldStartIndex: LongInt; NewStartIndex: LongInt; GroupCount: LongInt): Boolean;
var
  TempString: String;
  TempStringLen, StartPos, EndPos, EndIndex, IndexDiff: LongInt;
begin
  Result := False;
  if NewStartIndex < OldStartIndex then
  begin
    Result := True;
    EndIndex := NewStartIndex + GroupCount;
    StartPos := FLinePos[OldStartIndex];
    EndPos := FLinePos[OldStartIndex + GroupCount];
    TempStringLen := EndPos - StartPos - 2;
    SetString(TempString, (FOrigin + StartPos), TempStringLen);
    InsertString(NewStartIndex, TempString);
    OldStartIndex := OldStartIndex + GroupCount;
    DeleteGroup(OldStartIndex, GroupCount);
    TempString := '';
  end;

  if (NewStartIndex > OldStartIndex + GroupCount) then
  begin
    Result := True;
    EndIndex := NewStartIndex + GroupCount;
    StartPos := FLinePos[OldStartIndex];
    EndPos := FLinePos[OldStartIndex + GroupCount];
    TempStringLen := EndPos - StartPos - 2;
    SetString(TempString, (FOrigin + StartPos), TempStringLen);
    InsertString(NewStartIndex, TempString);
    DeleteGroup(OldStartIndex, GroupCount);
    TempString := '';
  end;
end; { MoveGroup }

function TPCharList.Remove(Item: String): Integer;
begin
  Result := IndexOf(Item);
  if Result <> -1 then Delete(Result);
end; { Remove }

procedure TPCharList.ScanForLines;
var
  Run: LongInt;
begin
  FLinePos.Capacity := 1;
  if fOrigin <> nil then
  begin
    Run := 0;
    while (fOrigin[Run] <> #0) do
      Case fOrigin[Run] of
        #10:
          begin
            inc(Run);
            FLinePos.Add(Run);
          end;

        #13:
          begin
            if fOrigin[Run + 1] = #10 then inc(Run, 2) else inc(Run);
            FLinePos.Add(Run);
          end;
      else inc(Run);
      end;
    if (not (fOrigin[Run - 1] in [#10, #13])) then FLinePos.Add(Run);
  end;
end; { ScanForLines }

procedure TPCharList.DeleteChars(CharCount: LongInt);
var
  DelPos: LongInt;
begin
  DelPos := FPosition + CharCount;
  if DelPos > fPCharSize then DelPos := fPCharSize;
  WriteTo(FPosition, DelPos, '');
  ScanForLines;
end; { DeleteChars }

function TPCharList.ReadChars(CharCount: LongInt): String;
begin
  while (FPosition + CharCount) > fPCharSize do dec(CharCount);
  SetString(Result, (FOrigin + FPosition), CharCount);
end; { ReadChars }

procedure TPCharList.WriteString(aString: String);
var
  DelPos: LongInt;
begin
  DelPos := FPosition + Length(aString);
  if DelPos > fPCharSize then DelPos := fPCharSize;
  WriteTo(FPosition, DelPos, aString);
  ScanForLines;
end; { WriteString }

end.

